(***  Simulation einer  Mehrprozessoranlage mit  3 Erzeugern und  ***)
(***  2 Verbrauchern,  die ber einen 10-elementigen Puffer ver-  ***)
(***  bunden sind.       (Literatur: Informatik III,  Langmaack)  ***)
(***  (c)  1985 by Claus Schnleber, Kiel; claus@schoenleber.com  ***)
(***  Die Semaphore sind nach Dijkstra implementiert, das heit,  ***)
(***  da fr einen Prozess  'warten'  ein Stillstehen bedeutet.  ***)

PROGRAM erzver (input,output,in1,in2,in3,out4,out5);
  CONST buffer_size      = 10;
  TYPE  binary_semaphore = boolean;
        semaphore        = 0..buffer_size;
        buffer_pointer   = 1..buffer_size;
        counter          = integer;
        status           = (waiting,working,completed);
  VAR   in1,in2,in3,out4,out5  : text;
        buff_in,buff_out       : buffer_pointer;
        process,index          : integer;
        current_number         : counter;
        mutex_write,mutex_read : binary_semaphore;
        full,empty             : semaphore;
        finished               : boolean;
        data_of_process        : ARRAY [1..5]           OF integer;
        program_counter        : ARRAY [1..9]           OF integer;
        eop                    : ARRAY [1..5]           OF boolean;
        pstat                  : ARRAY [1..5]           OF status;
        buffer                 : ARRAY [1..buffer_size] OF integer;

PROCEDURE initialize;
  VAR init : integer;
BEGIN
  writeln ('initializing...');
  finished       := false;
  mutex_read     := true;
  mutex_write    := true;
  full           := 0;
  empty          := buffer_size;
  buff_in        := 1;
  buff_out       := 1;
  process        := 0;
  current_number := 0;
  assign  (in1,'ERZ1.DAT');
  reset   (in1);
  assign  (in2,'ERZ2.DAT');
  reset   (in2);
  assign  (in3,'ERZ3.DAT');
  reset   (in3);
  assign  (out4,'VER4.DAT');
  rewrite (out4);
  assign  (out5,'VER5.DAT');
  rewrite (out5);
  FOR init := 1 TO buffer_size DO
    buffer [init] := 0;
  FOR init := 1 TO 5 DO
  BEGIN
    eop [init] := false;
    program_counter [init] := 1;
    pstat [init] := waiting
  END
END; (* of initialize *)

PROCEDURE wait (VAR wsema : semaphore);
BEGIN
  IF (wsema > 0) THEN
  BEGIN
    wsema := wsema - 1;
    pstat [process] := working;
    program_counter [process] := program_counter [process] + 1
  END
  ELSE
    pstat [process] := waiting
END; (* of wait *)

PROCEDURE wait_bin (VAR wbsema : binary_semaphore);
BEGIN
  IF wbsema THEN
  BEGIN
    wbsema := false;
    pstat [process] := working;
    program_counter [process] := program_counter [process] + 1
  END
  ELSE
    pstat [process] := waiting
END; (* of wait_bin *)

PROCEDURE signal (VAR ssema : semaphore);
BEGIN
  ssema := ssema + 1;
  program_counter [process] := program_counter [process] + 1
END;  (* of signal *)

PROCEDURE signal_bin (VAR sbsema : binary_semaphore);
BEGIN
  sbsema := true;
  program_counter [process] := program_counter [process] + 1
END; (* of signal_bin *)

FUNCTION next_process : integer;              (* Process-Supervisor *)
  VAR cpu : real;
BEGIN
  cpu := frac (9821 * cpu + 0.211327) * 5 + 1;
  next_process := trunc (cpu)
END; (* of next_process *)

PROCEDURE process_control;
  VAR process_number : integer;
BEGIN
  gotoxy (40,8);
  writeln ('Actual buffer contents :');
  gotoxy (40,10);
  FOR index := 1 TO 10 DO
    write (^Z'3',buffer [index]:2,^Z'2',' ');
  IF (buff_in <> buff_out) THEN
  BEGIN
    gotoxy (40,11);
    write ('                               ');
    gotoxy (buff_in * 3 + 37,11);
    write ('^I');
    gotoxy (buff_out * 3 + 37,11);
    write ('^O')
  END
  ELSE
  BEGIN
    gotoxy (40,11);
    write ('                               ');
    gotoxy (buff_in * 3 + 37,11);
    write ('^^')
  END;
  gotoxy (1,1);
  writeln ('Current number of processing : ',current_number:0);
  writeln ('Total process-control');
  writeln ('---------------------');
  FOR process_number := 1 TO 5 DO
  BEGIN
    writeln ('Process #',process_number:0);
    write ('  Status  : ');
    CASE pstat [process_number] OF
      working   : writeln ('working');
      waiting   : writeln ('waiting');
      completed : writeln ('completed')
    END;
    writeln ('  PC      : ',program_counter [process_number]:0);
    write ('  on line : ');
    IF process = process_number THEN
      writeln ('yes')
    ELSE
      writeln ('no ')
  END
END; (* process_control *)

PROCEDURE runtime;
BEGIN
  writeln ('processing...');
  WHILE NOT finished DO
  BEGIN
    current_number := current_number + 1;
    REPEAT
      process := next_process
    UNTIL NOT eop [process] OR finished;
    CASE process OF
      1 : CASE program_counter [1] OF
            1 : BEGIN
                  (* erzeuge *)
                  read (in1,data_of_process [1]);
                  program_counter [1] := program_counter [1] + 1
                END;
            2 : wait (empty);
            3 : wait_bin (mutex_write);
            4 : BEGIN
                  (* speichere *)
                  buffer [buff_in] := data_of_process [1];
                  IF buff_in = buffer_size THEN
                    buff_in := 1
                  ELSE
                    buff_in := buff_in + 1;
                  program_counter [1] := program_counter [1] + 1
                END;
            5 : signal_bin (mutex_write);
            6 : signal (full);
            7 : BEGIN
                  IF NOT eof (in1) THEN
                    program_counter [1] := 1
                  ELSE
                  BEGIN
                    eop [1] := true;
                    pstat [1] := completed
                  END
                END
          END;
      2 : CASE program_counter [2] OF
            1 : BEGIN
                  (* erzeuge *)
                  read (in2,data_of_process [2]);
                  program_counter [2] := program_counter [2] + 1
                END;
            2 : wait (empty);
            3 : wait_bin (mutex_write);
            4 : BEGIN
                  (* speichere *)
                  buffer [buff_in] := data_of_process [2];
                  IF buff_in = buffer_size THEN
                    buff_in := 1
                  ELSE
                    buff_in := buff_in + 1;
                  program_counter [2] := program_counter [2] + 1
                END;
            5 : signal_bin (mutex_write);
            6 : signal (full);
            7 : BEGIN
                  IF NOT eof (in2) THEN
                    program_counter [2] := 1
                  ELSE
                  BEGIN
                    eop [2] := true;
                    pstat [2] := completed
                  END
                END
          END;
      3 : CASE program_counter [3] OF
            1 : BEGIN
                  (* erzeuge *)
                  read (in3,data_of_process [3]);
                  program_counter [3] := program_counter [3] + 1
                END;
            2 : wait (empty);
            3 : wait_bin (mutex_write);
            4 : BEGIN
                  (* speichere *)
                  buffer [buff_in] := data_of_process [3];
                  IF buff_in = buffer_size THEN
                    buff_in := 1
                  ELSE
                    buff_in := buff_in + 1;
                  program_counter [3] := program_counter [3] + 1
                END;
            5 : signal_bin (mutex_write);
            6 : signal (full);
            7 : BEGIN
                  IF NOT eof (in3) THEN
                    program_counter [3] := 1
                  ELSE
                  BEGIN
                    eop [3] := true;
                    pstat [3] := completed
                  END
                END
          END;
      4 : CASE program_counter [4] OF
            1 : wait (full);
            2 : wait_bin (mutex_read);
            3 : BEGIN
                  (* lies *)
                  data_of_process [4] := buffer [buff_out];
                  buffer [buff_out] := 0;
                  IF buff_out = buffer_size THEN
                    buff_out := 1
                  ELSE
                    buff_out := buff_out + 1;
                  program_counter [4] := program_counter [4] + 1
                END;
            4 : signal_bin (mutex_read);
            5 : signal (empty);
            6 : BEGIN
                  (* verbrauche *)
                  writeln (out4,data_of_process [4]);
                  IF empty <> buffer_size THEN
                    program_counter [4] := 1
                  ELSE
                  BEGIN
                    eop [4] := true;
                    pstat [4] := completed
                  END
                END
          END;
      5 : CASE program_counter [5] OF
            1 : wait (full);
            2 : wait_bin (mutex_read);
            3 : BEGIN
                  (* lies *)
                  data_of_process [5] := buffer [buff_out];
                  buffer [buff_out] := 0;
                  IF buff_out = buffer_size THEN
                    buff_out := 1
                  ELSE
                    buff_out := buff_out + 1;
                  program_counter [5] := program_counter [5] + 1
                END;
            4 : signal_bin (mutex_read);
            5 : signal (empty);
            6 : BEGIN
                  (* verbrauche *)
                  writeln (out5,data_of_process [5]);
                  IF empty <> buffer_size THEN
                    program_counter [5] := 1
                  ELSE
                  BEGIN
                    eop [5] := true;
                    pstat [5] := completed
                  END
                END
          END
    END;
    finished := eop [1] AND eop [2] AND eop [3] AND (eop [4] OR eop [5]);
    process_control;
  END;
  close (out4);
  close (out5)
END; (* of runtime *)

(*** MAIN of Simulation ***)

BEGIN
  clrscr;
  initialize;
  runtime
END.